"""Tests for client module."""
from unittest import TestCase
from unittest.mock import Mock, call, patch, sentinel

import grpc
import grpc._channel
from google.protobuf.empty_pb2 import Empty
from testfixtures import LogCapture, StringComparison

from frgal.client import GrpcClient
from frgal.decoder import GrpcDecoder
from frgal.exceptions import ServiceNotFound, UnknownError
from frgal.future import FutureCallWrapper

from .proto.example_pb2 import Reply, SimpleMessage

try:
    from grpc._channel import _MultiThreadedRendezvous as _Rendezvous
except ImportError:
    from grpc._channel import _Rendezvous


class TestGrpcClient(TestCase):
    """Test GrpcClient."""

    def setUp(self):
        self.grpc_module_mock = Mock(spec=('coffeeStub', ))

        self.log_handler = LogCapture()
        self.addCleanup(self.log_handler.uninstall)

    def test_connect_insecure(self):
        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc)])

    def test_connect_secure(self):
        with patch('frgal.client.grpc.secure_channel', return_value=sentinel.channel) as channel_mock:
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock],
                                credentials=sentinel.credentials)

        self.assertEqual(client.channel, sentinel.channel)
        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc, sentinel.credentials)])

    def test_get_stub(self):
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_cache(self):
        # Test stubs are cached
        self.grpc_module_mock.coffeeStub.return_value = sentinel.coffee_stub

        with patch('frgal.client.grpc.insecure_channel', return_value=sentinel.channel):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        # Get stub once
        client.get_stub('coffee')
        # Get stub twice
        self.assertEqual(client.get_stub('coffee'), sentinel.coffee_stub)
        # Check it was created only once
        self.assertEqual(self.grpc_module_mock.mock_calls, [call.coffeeStub(sentinel.channel)])

    def test_get_stub_unknown(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, sentinel.decoder, [self.grpc_module_mock])

        with self.assertRaisesRegex(ServiceNotFound, 'Service teaStub could not be found'):
            client.get_stub('tea')

    def test_call(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.future.return_value = Mock(spec=['result'])
        stub_mock.make.future.return_value.result.return_value = reply

        self.assertEqual(client.call('coffee', 'make', Empty()), 'Espresso!')

    def test_call_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.exception.tea_pot_error.CopyFrom(Reply.Exception.TeaPotError())
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.future.return_value = Mock(spec=['result'])
        stub_mock.make.future.return_value.result.return_value = reply

        with self.assertRaisesRegex(UnknownError, 'tea_pot_error'):
            client.call('coffee', 'make', Empty())

    def test_call_empty(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Empty()
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.future.return_value = Mock(spec=['result'])
        stub_mock.make.future.return_value.result.return_value = reply

        self.assertIsNone(client.call('coffee', 'make', Empty()))

    def test_call_grpc_unavailable(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        future_mock = Mock(autospec=grpc.Future)
        future_mock.result.side_effect = [error, error, error, reply]
        stub_mock.make.future.return_value = future_mock

        with patch('frgal.client.grpc.insecure_channel') as channel_mock:
            self.assertEqual(client.call('coffee', 'make', Empty()), 'Espresso!')

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc), call(sentinel.netloc), call(sentinel.netloc)])
        self.assertEqual(future_mock.mock_calls, [
            call.result(None), call.cancel(),
            call.result(None), call.cancel(),
            call.result(None), call.cancel(),
            call.result(None),
        ])
        self.assertEqual(len(self.log_handler.records), 5)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertEqual(self.log_handler.records[1].getMessage(),
                         'Connection to gRPC service coffee failed. Trying to reconnect.')
        self.assertEqual(self.log_handler.records[2].getMessage(),
                         'Connection to gRPC service coffee failed. Trying to reconnect.')
        self.assertEqual(self.log_handler.records[3].getMessage(),
                         'Connection to gRPC service coffee failed. Trying to reconnect.')
        self.assertRegex(self.log_handler.records[4].getMessage(), r'\[\w{4}\] make returned .*')

    def test_call_grpc_unavailable_too_many_times(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNAVAILABLE, "Connection reset by peer")
        error = _Rendezvous(state, None, None, None)
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        future_mock = Mock(autospec=grpc.Future)
        future_mock.result.side_effect = error
        stub_mock.make.future.return_value = future_mock

        with patch('frgal.client.grpc.insecure_channel') as channel_mock:
            response = client.call('coffee', 'make', Empty(), blocking=False)
            with self.assertRaises(grpc.RpcError):
                response.result(max_retries=1)

        self.assertEqual(channel_mock.mock_calls, [call(sentinel.netloc)])
        self.assertEqual(future_mock.mock_calls, [call.result(None), call.cancel(), call.result(None)])
        self.assertEqual(len(self.log_handler.records), 3)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertEqual(self.log_handler.records[1].getMessage(),
                         'Connection to gRPC service coffee failed. Trying to reconnect.')
        self.assertRegex(self.log_handler.records[2].getMessage(), r'\[\w{4}\] make failed with .*')

    def test_call_grpc_other_error(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        state = grpc._channel._RPCState((), '', '', grpc.StatusCode.UNKNOWN, "Unknown error")
        error = _Rendezvous(state, None, None, None)
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        future_mock = Mock(spec=['result'])
        future_mock.result.side_effect = error
        stub_mock.make.future.return_value = future_mock

        with self.assertRaisesRegex(_Rendezvous, 'status = StatusCode.UNKNOWN'):
            client.call('coffee', 'make', Empty())

        self.assertEqual(len(self.log_handler.records), 2)
        self.assertRegex(self.log_handler.records[0].getMessage(), r'\[\w{4}\] make(.*)')
        self.assertIn('Rendezvous of RPC that terminated with:', self.log_handler.records[1].getMessage())

    def test_call_non_blocking(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        future_mock = Mock(autospec=grpc.Future)
        future_mock.result.return_value = reply
        future_mock.code.return_value = 'OK'
        stub_mock.make.future.return_value = future_mock

        result = client.call('coffee', 'make', Empty(), blocking=False)
        self.assertIsInstance(result, FutureCallWrapper)
        self.assertEqual(result.result(), 'Espresso!')
        self.assertEqual(result.code(), 'OK')

    def test_call_modify_result(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.future.return_value = Mock(spec=['result'])
        stub_mock.make.future.return_value.result.return_value = reply

        def modify_result(result):
            return '{} Shaken, not stirred.'.format(result)

        self.assertEqual(
            client.call('coffee', 'make', Empty(), modify_result=modify_result),
            'Espresso! Shaken, not stirred.'
        )

    def test_call_logging(self):
        with patch('frgal.client.grpc.insecure_channel'):
            client = GrpcClient(sentinel.netloc, GrpcDecoder(), [self.grpc_module_mock])
        reply = Reply()
        reply.data.result = 'Espresso!'
        stub_mock = self.grpc_module_mock.coffeeStub.return_value
        stub_mock.make.future.return_value = Mock(spec=['result'])
        stub_mock.make.future.return_value.result.return_value = reply

        with self.assertLogs('frgal', level='DEBUG') as logs:
            self.assertEqual(client.call('coffee', 'make', SimpleMessage(value='10')), 'Espresso!')
            self.assertEqual(logs.output, [
                StringComparison(r'DEBUG:frgal\.client:\[.{4}\] make\(value: "10"\)'),
                StringComparison(r'DEBUG:frgal\.future:\[.{4}\] make returned data \{ result: "Espresso!" \}'),
            ])
